/* Generates a maze for sauer in a format suitable for txt2oct */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include "takehana.h"

#define SUBDIV 4
#define WIDTH (1 << SUBDIV)
#define HEIGHT (1 << SUBDIV)

typedef struct CellList {
  int x, y;
  int members;/* only correct for first entry */
  struct CellList *prev, *next;
} CellList;

CellList *frontier;

/* The maze
   # - wall
   . - walkable
   , - exposed but not finished with
   ? - unexposed */
unsigned char grid[HEIGHT][WIDTH];

/* Adds the given co-ords to the start of the given list */
CellList *append(CellList **list, int x, int y) {
  CellList *node;

  node = malloc(sizeof(CellList));
  node->x = x;
  node->y = y;
  node->prev = NULL;
  node->next = *list;
  node->members = 1 + (*list ? (*list)->members : 0);

  if(*list) (*list)->prev = node;

  *list = node;

  return node;
}

/* Returns the nth member from the given list, or NULL if there aren't enough.
   Warning: this takes O(n) time */
CellList *member(CellList *list, int n) {
  CellList *node;

  if(n < 0) return NULL;

  for(node = list; n && node; n--) {
    node = node->next;
  }

  return node;
}

/* Removes the given node from the given list. The list pointer must be given
   because we need to decrement the member count. Note that it is not checked
   that node is actually in list. Be consistent or the (*list)->members field
   will be wrong. It is possible that *list will be NULL upon return, if e.g.
   the list becomes empty */
void delete(CellList **list, CellList *node) {
  int members = (*list)->members;

  if(node->prev) node->prev->next = node->next;
  else *list = node->next;

  if(node->next) node->next->prev = node->prev;
  free(node);

  if(*list) (*list)->members = members - 1;
}

/* Makes grid[y][x] empty, and adds new cells to the list of exposed cells.
   Does not remove this cell from frontier */
void carve(int x, int y) {
  int xx, yy;

  grid[y][x] = '.';

  for(yy = -1; yy <= 1; yy++) {
    for(xx = -1; xx <= 1; xx++) {
      if(xx == 0 && yy == 0) continue;
      if(xx != 0 && yy != 0) continue;

      if(x + xx >= 0 && x + xx < WIDTH && y + yy >= 0 && y + yy < HEIGHT
         && grid[y + yy][x + xx] == '?') {
        grid[y + yy][x + xx] = ',';
        append(&frontier, x + xx, y + yy);
      }
    }
  }
}

/* Makes the given square non-walkable */
void harden(int x, int y) {
  grid[y][x] = '#';
}

/* returns 1 if it is possible to carve out the given co-ordinates and 0
   otherwise. */
int cancarve(int x, int y) {
  int touches = 0;

  /* this square can be made walkable if and only if it touches exactly one
     other walkable tile */
  if(x != 0 && grid[y][x-1] == '.') touches++;
  if(x != WIDTH-1 && grid[y][x+1] == '.') touches++;
  if(y != 0 && grid[y-1][x] == '.') touches++;
  if(y != HEIGHT-1 && grid[y+1][x] == '.') touches++;
    
  if(touches == 1) return 1;
  else return 0;
}

void output(int x, int y, int z, int subdiv) {
  int xx, yy, zz;

  for(zz = 0; zz < 2; zz++) {
    for(yy = 0; yy < 2; yy++) {
      for(xx = 0; xx < 2; xx++) {
        if(subdiv < SUBDIV-1) {/* need to split the world mode */
          printf("OCTSAV_CHILDREN\n");
          output((x + xx) * 2, (y + yy) * 2, (z + zz) * 2, subdiv + 1);
        } else {/* output solid or empty cubes */
          if(z + zz == HEIGHT/2 && grid[y + yy][x + xx] == '.')
            printf("OCTSAV_EMPTY 0 0 0 0 0 0\n");
          else if(z + zz > HEIGHT/2)
            printf("OCTSAV_EMPTY 0 0 0 0 0 0\n");
          else printf("OCTSAV_SOLID 3 4 5 6 7 8\n");
         }
      }
    }
  }
}

int main(int argc, char **argv) {
  int x, y;
  int n;
  CellList *choice;
  int seed;

  if(argc == 2) {
    seed = atoi(argv[1]);
  } else {
    seed = time(NULL);
  }

  fprintf(stderr, "Seeding RNG to %d.\n", seed);
  srand(seed);

  memset(grid, '?', WIDTH * HEIGHT);

  /* carve out the starting square */
  carve(0, 0);
  carve(WIDTH-1, HEIGHT-1);
  
  /* while there are still non-determined squares to be mined */
  while(frontier) {
    /* choose a random square from the frontier */
    n = rand() % frontier->members;
    choice = member(frontier, n);

    /* mine out or harden */
    if(cancarve(choice->x, choice->y)) carve(choice->x, choice->y);
    else harden(choice->x, choice->y);

    /* and this square is no longer non-determined */
    delete(&frontier, choice);
  }

  /* print out the maze */
  for(y = 0; y < HEIGHT; y++) {
    for(x = 0; x < WIDTH; x++) {
      if(grid[y][x] == '?') grid[y][x] = '#';/* non-exposed square */
      fputc(grid[y][x], stderr);
    }
    fputc('\n', stderr);
  }

  /* print out maze in txt2ogz format */
  output(0, 0, 0, 0);

  return 0;
}
